from __future__ import division

import copy
import time

import numpy as np
from _AwminTV import AwminTV
from _minTV import minTV
from tigre.algorithms.single_pass_algorithms import FDK
from tigre.utilities.Atb import Atb
from tigre.utilities.Ax import Ax
from tigre.utilities.im3Dnorm import im3DNORM
from tigre.utilities.init_multigrid import init_multigrid
from tigre.utilities.order_subsets import order_subsets
from tigre.utilities.Measure_Quality import Measure_Quality as MQ
from tigre.utilities.gpu import GpuIds

"""
This module is where the umbrella class IterativeReconAlg is located
which is the umbrella class to all the other algorithms apart from
the single pass type algorithms.
"""

# coding: utf8


if hasattr(time, "perf_counter"):
    default_timer = time.perf_counter
else:
    default_timer = time.clock


class IterativeReconAlg(object):
    """
    Parameters
    ----------
    :param proj: (np.ndarray, dtype=np.float32)
    Input data, shape = (geo.nDector, nangles)

    :param geo: (tigre.geometry)
    Geometry of detector and image (see examples/Demo code)

    :param angles: (np.ndarray , dtype=np.float32)
    angles of projection, shape = (nangles,3)

    :param niter: (int)
    number of iterations for reconstruction algorithm

    :param kwargs: (dict)
    optional parameters

    Keyword Arguments
    -----------------
    :keyword blocksize: (int)
        number of angles to be included in each iteration
        of proj and backproj for OS_SART
    :keyword lmbda: (np.float64)
        Sets the value of the hyperparameter.

    :keyword lmbda_red: (np.float64)
        Reduction of lmbda every iteration
        lmbda=lmbda_red*lmbda. Default is 0.99

    :keyword init: (str)
        Describes different initialization techniques.
               None      : Initializes the image to zeros (default)
              "FDK"      : intializes image to FDK reconstrucition

    :keyword verbose:  (Boolean)
        Feedback print statements for algorithm progress
        default=True

    :keyword OrderStrategy : (str)
        Chooses the subset ordering strategy. Options are:
                 "ordered"        : uses them in the input order, but
                                    divided
                 "random"         : orders them randomply

    :keyword tviter: (int)
        For algorithms that make use of a tvdenoising step in their
        iterations. This includes:

            OS_SART_TV
            ASD_POCS
            AWASD_POCS
            FISTA

    :keyword tvlambda: (float)
        For algorithms that make use of a tvdenoising step in their
        iterations.

            OS_SART_TV
            FISTA

    Usage
    --------
    >>> import numpy as np
    >>> import tigre
    >>> import tigre.algorithms as algs
    >>> from tigre.demos.Test_data import data_loader
    >>> geo = tigre.geometry(mode='cone',default_geo=True,
    >>>                         nVoxel=np.array([64,64,64]))
    >>> angles = np.linspace(0,2*np.pi,100)
    >>> src_img = data_loader.load_head_phantom(geo.nVoxel)
    >>> proj = tigre.Ax(src_img,geo,angles)
    >>> output = algs.iterativereconalg(proj,geo,angles,niter=50
    >>>                                 blocksize=20)

    tigre.demos.run() to launch ipython notebook file with examples.


    --------------------------------------------------------------------
    This file is part of the TIGRE Toolbox

    Copyright (c) 2015, University of Bath and
                        CERN-European Organization for Nuclear Research
                        All rights reserved.

    License:            Open Source under BSD.
                        See the full license at
                        https://github.com/CERN/TIGRE/license.txt

    Contact:            tigre.toolbox@gmail.com
    Codes:              https://github.com/CERN/TIGRE/
    --------------------------------------------------------------------
    Coded by:          MATLAB (original code): Ander Biguri
                       PYTHON : Reuben Lindroos
    """

    def __init__(self, proj, geo, angles, niter, **kwargs):

        self.proj = proj
        self.angles = angles
        self.geo = geo
        self.niter = niter

        self.geo.check_geo(angles)

        options = dict(
            blocksize=20,
            lmbda=1,
            lmbda_red=1,
            OrderStrategy=None,
            Quameasopts=None,
            init=None,
            verbose=True,
            noneg=True,
            computel2=False,
            dataminimizing="art_data_minimizing",
            name="Iterative Reconstruction",
            sup_kw_warning=False,
            gpuids=None,
            niter_outer=4,
            restart=True
        )
        allowed_keywords = [
            "V",
            "W",
            "log_parameters",
            "angleblocks",
            "angle_index",
            "alpha",
            "alpha_red",
            "rmax",
            "maxl2err",
            "delta",
            "regularisation",
            "tviter",
            "tvlambda",
            "hyper",
            "fista_p",
            "fista_q",
            "niter_outer"
        ]
        self.__dict__.update(options)
        self.__dict__.update(**kwargs)
        for kw in kwargs.keys():
            if kw not in options and (kw not in allowed_keywords):
                if self.verbose:
                    if not kwargs.get("sup_kw_warning"):
                        # Note: might not want this warning (typo checking).
                        print(
                            "Warning: "
                            + kw
                            + " not recognised as default parameter for instance of IterativeReconAlg."  # noqa: E501
                        )
        if self.angles.ndim == 1:
            a1 = self.angles
            a2 = np.zeros(self.angles.shape[0], dtype=np.float32)
            self.angles = np.vstack((a1, a2, a2)).T
        if not all([hasattr(self, "angleindex"), hasattr(self, "angleblocks")]):
            self.set_angle_index()
        if not hasattr(self, "W"):
            self.set_w()
        if not hasattr(self, "V"):
            self.set_v()
        if not hasattr(self, "res"):
            self.set_res()
        if self.verbose:
            self.tic = 0    # preparation for _estimate_time_until_completion()
        # make it list
        if self.Quameasopts is not None:
            self.Quameasopts = (
                [self.Quameasopts] if isinstance(self.Quameasopts, str) else self.Quameasopts
            )
            setattr(self, "lq", np.zeros([len(self.Quameasopts), niter]))  # quameasoptslist
        else:
            setattr(self, "lq", np.zeros([0, niter]))  # quameasoptslist
        setattr(self, "l2l", np.zeros([1, niter]))  # l2list

    def set_w(self):
        """
        Calculates value of W if this is not given.
        :return: None
        """
        geox = copy.deepcopy(self.geo)
        geox.sVoxel[1:] = geox.sVoxel[1:] * 1.1  # a bit larger to avoid zeros in projections
        geox.sVoxel[0] = max(geox.sDetector[0], geox.sVoxel[0])

        geox.nVoxel = np.array([2, 2, 2])
        geox.dVoxel = geox.sVoxel / geox.nVoxel
        W = Ax(
            np.ones(geox.nVoxel, dtype=np.float32), geox, self.angles, "Siddon", gpuids=self.gpuids
        )
        W[W <= min(self.geo.dVoxel / 2)] = np.inf
        W = 1.0 / W
        setattr(self, "W", W)

    def set_v(self):
        """
        Computes value of V parameter if this is not given.
        :return: None
        """
        block_count = len(self.angleblocks)
        geo = self.geo
        V = np.ones((block_count, geo.nVoxel[1], geo.nVoxel[2]), dtype=np.float32)

        for i in range(block_count):
            if geo.mode != "parallel":

                geox = copy.deepcopy(self.geo)
                geox.angles = self.angleblocks[i]

                geox.DSD = geo.DSD[self.angle_index[i]]
                geox.DSO = geo.DSO[self.angle_index[i]]
                geox.offOrigin = geo.offOrigin[self.angle_index[i], :]
                geox.offDetector = geo.offDetector[self.angle_index[i], :]
                geox.rotDetector = geo.rotDetector[self.angle_index[i], :]
                geox.COR = geo.COR[self.angle_index[i]]

                # shrink the volume size to avoid zeros in backprojection
                geox.sVoxel = (
                    geox.sVoxel * np.max(geox.sVoxel[1:] / np.linalg.norm(geox.sVoxel[1:])) * 0.9
                )
                geox.dVoxel = geox.sVoxel / geox.nVoxel
                proj_one = np.ones(
                    (len(self.angleblocks[i]), geo.nDetector[0], geo.nDetector[1]), dtype=np.float32
                )
                V[i] = Atb(proj_one, geox, self.angleblocks[i], "FDK", gpuids=self.gpuids).mean(
                    axis=0
                )

            else:
                V[i] *= len(self.angleblocks[i])
        V[V==0.0] = np.inf       

        self.V = V

    def set_res(self):
        """
        Calulates initial value for res if this is not given.
        :return: None
        """
        self.res = np.zeros(self.geo.nVoxel, dtype=np.float32)
        init = self.init
        verbose = self.verbose
        if isinstance(init, str):
            if init == "multigrid":
                if verbose:
                    print("init multigrid in progress...")
                    print("default blocksize=1 for init_multigrid(OS_SART)")
                self.res = init_multigrid(self.proj, self.geo, self.angles, alg="SART")
                if verbose:
                    print("init multigrid complete.")
            if init == "FDK":
                self.res = FDK(self.proj, self.geo, self.angles)

        elif isinstance(init, np.ndarray):
            if (self.geo.nVoxel == init.shape).all():
                self.res = init
            else:
                raise ValueError("wrong dimension of array for initialisation")
        elif init is not None:
            raise ValueError("wrong value for initialisation")

    def set_angle_index(self):
        """
        sets angle_index and angleblock if this is not given.
        :return: None
        """
        self.angleblocks, self.angle_index = order_subsets(
            self.angles, self.blocksize, self.OrderStrategy
        )

    def run_main_iter(self):
        """
        Goes through the main iteration for the given configuration.
        :return: None
        """
        Quameasopts = self.Quameasopts

        for i in range(self.niter):

            res_prev = None
            if Quameasopts is not None:
                res_prev = copy.deepcopy(self.res)
            if self.verbose:
                self._estimate_time_until_completion(i)

            getattr(self, self.dataminimizing)()
            self.error_measurement(res_prev, i)

    def art_data_minimizing(self):
        geo = copy.deepcopy(self.geo)

        for j in range(len(self.angleblocks)):

            if self.blocksize == 1:
                angle = np.array([self.angleblocks[j]], dtype=np.float32)
                angle_indices = np.array([self.angle_index[j]], dtype=np.int32)

            else:
                angle = self.angleblocks[j]
                angle_indices = self.angle_index[j]
                # slice parameters if needed
            geo.offOrigin = self.geo.offOrigin[angle_indices]
            geo.offDetector = self.geo.offDetector[angle_indices]
            geo.rotDetector = self.geo.rotDetector[angle_indices]
            geo.DSD = self.geo.DSD[angle_indices]
            geo.DSO = self.geo.DSO[angle_indices]

            self.update_image(geo, angle, j)

            if self.noneg:
                self.res = self.res.clip(min=0)

    def minimizeTV(self, res_prev, dtvg):
        if self.gpuids is None:
            self.gpuids = GpuIds()
        return minTV(res_prev, dtvg, self.numiter_tv, self.gpuids)

    def minimizeAwTV(self, res_prev, dtvg):
        if self.gpuids is None:
            self.gpuids = GpuIds()
        return AwminTV(res_prev, dtvg, self.numiter_tv, self.delta, self.gpuids)

    def error_measurement(self, res_prev, iter):
        if self.Quameasopts is not None:
            self.lq[:, iter] = MQ(self.res, res_prev, self.Quameasopts)
        if self.computel2:
            # compute l2 borm for b-Ax
            errornow = im3DNORM(
                self.proj - Ax(self.res, self.geo, self.angles, "Siddon", gpuids=self.gpuids), 2
            )
            self.l2l[0, iter] = errornow

    def update_image(self, geo, angle, iteration):
        """
        VERBOSE:
         for j in range(angleblocks):
             angle = np.array([alpha[j]], dtype=np.float32)
             proj_err = proj[angle_index[j]] - Ax(res, geo, angle, 'Siddon')
             weighted_err = W[angle_index[j]] * proj_err
             backprj = Atb(weighted_err, geo, angle, 'FDK')
             weighted_backprj = 1 / V[angle_index[j]] * backprj
             res += weighted_backprj
             res[res<0]=0

        :return: None
        """

        ang_index = self.angle_index[iteration].astype(np.int)

        self.res += (
            self.lmbda
            * 1.0
            / self.V[iteration]
            * Atb(
                self.W[ang_index]
                * (self.proj[ang_index] - Ax(self.res, geo, angle, "Siddon", gpuids=self.gpuids)),
                geo,
                angle,
                "FDK",
                gpuids=self.gpuids,
            )
        )

    def getres(self):
        return self.res

    def geterrors(self):
        if self.computel2:
            return np.concatenate((self.l2l, self.lq), axis=0)
        else:
            return self.lq

    def __str__(self):
        parameters = []
        for item in self.__dict__:
            if item == "geo":
                pass
            elif hasattr(self.__dict__.get(item), "shape"):
                if self.__dict__.get(item).ravel().shape[0] > 100:
                    parameters.append(item + " shape: " + str(self.__dict__.get(item).shape))
            else:
                parameters.append(item + ": " + str(self.__dict__.get(item)))

        return "\n".join(parameters)

    def _estimate_time_until_completion(self, iter):
        if iter == 0:
            print(str(self.name).upper() + " " + "algorithm in progress.")
            self.tic = default_timer()
        if iter == 1:
            toc = default_timer()

            remaining_time = (self.niter - 1) * (toc - self.tic)
            seconds = int(remaining_time)
            print(
                "Estimated time until completion : "
                + time.strftime("%H:%M:%S", time.gmtime(seconds))
            )


def decorator(IterativeReconAlg, name=None, docstring=None):  # noqa: N803
    """
    Calls run_main_iter when parameters are given to it.

    :param IterativeReconAlg: obj, class
        instance of IterativeReconAlg
    :param name: str
        for name of func
    :param docstring: str
        other documentation that may need to be included from external source.
    :return: func

    Examples
    --------
    >>> import tigre
    >>> from tigre.demos.Test_data.data_loader import load_head_phantom
    >>> geo = tigre.geometry_defaut(high_resolution=False)
    >>> src = load_head_phantom(number_of_voxels=geo.nVoxel)
    >>> proj = Ax(src,geo,angles)
    >>> angles = np.linspace(0,2*np.pi,100)
    >>> iterativereconalg = decorator(IterativeReconAlg)
    >>> output = iterativereconalg(proj,geo,angles, niter=50)

    """

    def iterativereconalg(proj, geo, angles, niter, **kwargs):

        geo.check_geo(angles)
        alg = IterativeReconAlg(proj, geo, angles, niter, **kwargs)
        if name is not None:
            alg.name = name
        alg.run_main_iter()
        if alg.computel2 or alg.Quameasopts is not None:
            return alg.getres(), alg.geterrors()
        else:
            return alg.getres()

    if docstring is not None:
        iterativereconalg.__doc__ = docstring + IterativeReconAlg.__doc__
    else:
        iterativereconalg.__doc__ = IterativeReconAlg.__doc__
    if name is not None:
        iterativereconalg.__name__ = name

    return iterativereconalg


iterativereconalg = decorator(IterativeReconAlg)



import numpy as np
import _RandomNumberGenerator as RNG

def add(projections, Gaussian=None, Poisson=None):

    if Poisson is not None:
        if not np.isscalar(Poisson):
            raise ValueError(
                "Poisson value should be an scalar, is " + str(type(Poisson)) + " instead."
            )
    else:
        Poisson = np.ceil(np.log2(np.max(np.abs(projections))))  # nextpow2
    if Gaussian is not None:
        if not isinstance(Gaussian, np.ndarray):
            raise ValueError(
                "Gaussian value should be an array, is " + str(type(Gaussian)) + " instead."
            )
        if Gaussian.shape != (2,):
            raise ValueError("Gaussian shape should be 1x2, is " + str(Gaussian.shape) + "instead.")

    max_proj = np.max(projections)
    projections = Poisson * np.exp(-projections / max_proj)

    projections = RNG.add_noise(projections, Gaussian[0], Gaussian[1])

    projections = -np.log(projections / Poisson) * max_proj
    projections = np.float32(projections)
    return projections



import copy

import numpy as np
from _Atb import _Atb_ext

from .gpu import GpuIds


def Atb(projections, geo, angles, backprojection_type="FDK", **kwargs):
    if projections.dtype != np.float32:
        raise TypeError("Input data should be float32, not " + str(projections.dtype))
    if not np.isreal(projections).all():
        raise ValueError("Complex types not compatible for back projection.")
    geox = copy.deepcopy(geo)
    geox.check_geo(angles)
    """
    Here we cast all values in geo to single point precision float. This way we know what behaviour
    to expect from pytigre to Cuda and can change single parameters accordingly.
    """
    geox.cast_to_single()
    # geox.checknans()

    if projections.shape != (geox.angles.shape[0], geox.nDetector[0], geox.nDetector[1]):
        raise ValueError(
            "Expected data shape: "
            + str((geox.angles.shape[0], geox.nDetector[0], geox.nDetector[1]))
            + " not compatible with: "
            + str(projections.shape)
        )

    if kwargs.get("gpuids", None) is None:
        gpuids = GpuIds()
    else:
        gpuids = kwargs["gpuids"]

    return _Atb_ext(projections, geox, geox.angles, backprojection_type, geox.mode, gpuids=gpuids)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from .art_family_algorithms import sart
from .art_family_algorithms import sirt
from .art_family_algorithms import ossart
from .art_family_algorithms import sart_tv
from .art_family_algorithms import ossart_tv
from .ista_algorithms import fista
from .ista_algorithms import ista
from .iterative_recon_alg import iterativereconalg
from .krylov_subspace_algorithms import cgls
from .krylov_subspace_algorithms import lsqr
from .krylov_subspace_algorithms import hybrid_lsqr
from .krylov_subspace_algorithms import lsmr
from .krylov_subspace_algorithms import irn_tv_cgls
from .krylov_subspace_algorithms import hybrid_flsqr_tv
from .krylov_subspace_algorithms import ab_gmres
from .krylov_subspace_algorithms import ba_gmres
from .pocs_algorithms import asd_pocs
from .pocs_algorithms import os_asd_pocs
from .pocs_algorithms import awasd_pocs
from .pocs_algorithms import os_awasd_pocs
from .pocs_algorithms import pcsd
from .pocs_algorithms import aw_pcsd
from .pocs_algorithms import os_pcsd
from .pocs_algorithms import os_aw_pcsd
from .single_pass_algorithms import fdk
from .single_pass_algorithms import fbp
from .statistical_algorithms import mlem

__all__ = [
    "sart",
    "sirt",
    "ossart",
    "sart_tv",
    "ossart_tv",
    "iterativereconalg",
    "FDK",
    "asd_pocs",
    "os_asd_pocs",
    "awasd_pocs",
    "os_awasd_pocs",
    "pcsd",
    "os_pcsd",
    "aw_pcsd",
    "os_aw_pcsd",
    "fbp",
    "cgls",
    "lsqr",
    "lsmr",
    "fista",
    "ista",
    "mlem",
]



from __future__ import division

import copy
import numpy as np
from tigre.algorithms.iterative_recon_alg import IterativeReconAlg
from tigre.algorithms.iterative_recon_alg import decorator
from tigre.utilities.Atb import Atb
from tigre.utilities.Ax import Ax


class MLEM(IterativeReconAlg):  # noqa: D101
    __doc__ = (
        " MLEM solves the CBCT problem using the maximum likelihood expectation maximization\n"
        " algorithm\n"
        " \n"
        "  MLEM(PROJ,GEO,ANGLES,NITER,INIT) solves the reconstruction problem\n"
        "  using the projection data PROJ taken over ALPHA angles, corresponding\n"
        "  to the geometry descrived in GEO, using NITER iterations. INIT specifies\n"
        "  starting image, defaut: None (flat image value=1)"
    ) + IterativeReconAlg.__doc__

    def __init__(self, proj, geo, angles, niter, **kwargs):
        # Don't precompute V and W.
        kwargs.update(dict(W=None, V=None))
        kwargs.update(dict(blocksize=angles.shape[0]))
        IterativeReconAlg.__init__(self, proj, geo, angles, niter, **kwargs)

        if self.init is None:
            self.res += 1.0

        self.W = Atb(np.ones(proj.shape, dtype=np.float32), geo, angles, backprojection_type="matched", gpuids=self.gpuids)
        self.W[self.W <= 0.0] = np.inf

    # Overide
    def run_main_iter(self):
        self.res[self.res < 0.0] = 0.0
        for i in range(self.niter):
            if self.Quameasopts is not None:
                res_prev = copy.deepcopy(self.res)
            self._estimate_time_until_completion(i)

            den = Ax(self.res, self.geo, self.angles, "interpolated", gpuids=self.gpuids)
            den[den == 0.0] = np.inf
            auxmlem = self.proj / den

            # update
            img = Atb(auxmlem, self.geo, self.angles, backprojection_type="matched", gpuids=self.gpuids) / self.W  

            self.res = self.res * img
            self.res[self.res < 0.0] = 0.0
            if self.Quameasopts is not None:
                self.error_measurement(res_prev, i)


mlem = decorator(MLEM, name="mlem")



from __future__ import division
from __future__ import print_function

import warnings

import numpy as np


def parkerweight(proj, geo, angles, q):
    start = -geo.sDetector[0] / 2 + geo.dDetector[0] / 2
    stop = geo.sDetector[0] / 2 - geo.dDetector[0] / 2
    step = geo.dDetector[0]
    alpha = np.arctan(np.arange(start, stop + step, step) / geo.DSD)
    alpha = -alpha
    delta = abs(alpha[0] - alpha[-1]) / 2
    totangles = np.cumsum(np.diff(angles))
    totangles = totangles[-1]

    if totangles < 2 * np.pi:
        warnings.warn(
            "Computing Parker weigths for scanning angle equal or bigger than 2*pi "
            "Consider disabling Parker weigths."
        )
    if totangles < np.pi + 2 * delta:
        warnings.warn(
            "Scanning angles smaller than pi+cone_angle. This is limited angle tomgraphy, \n"
            "there is nosufficient data, thus weigthing for data redundancy is not required."
        )
    epsilon = max(totangles - (np.pi + 2 * delta), 0)

    # for i in range(proj.shape[0]):
    for i in range(33):
        beta = angles[i]
        w = 0.5 * (
            s_function(beta / b_subf(alpha, delta, epsilon, q) - 0.5)
            + s_function(
                (beta - 2 * delta + 2 * alpha - epsilon) / b_subf(alpha, delta, epsilon, q) + 0.5
            )  # noqa: E501
            - s_function((beta - np.pi + 2 * alpha) / b_subf(-alpha, delta, epsilon, q) - 0.5)
            - s_function(
                (beta - np.pi - 2 * delta - epsilon) / b_subf(-alpha, delta, epsilon, q) + 0.5
            )  # noqa: E501
        )
        proj[i] *= w

    return proj


def s_function(abeta):
    w = np.zeros(abeta.shape)
    w[abeta <= -0.5] = 0
    w[abs(abeta) < 0.5] = 0.5 * (1 + np.sin(np.pi * abeta[abs(abeta) < 0.5]))
    w[abeta >= 0.5] = 1
    return w


# b_function = B
def b_function(alpha, delta, epsilon):
    return 2 * delta - 2 * alpha + epsilon


# b_subf = b
def b_subf(alpha, delta, epsilon, q):
    return q * b_function(alpha, delta, epsilon)



from __future__ import print_function
from __future__ import with_statement

import os
import math
import numpy
from tqdm import tqdm

from PIL import Image
from configparser import ConfigParser
from tigre.utilities.geometry import Geometry


def NikonDataLoader(filepath, **kwargs):
    # NikonDataLoader(filepath) Loads Nikon uCT datasets into TIGRE standard
    #
    # NikonDataLoader(filepath, OPT=VAL, ...) uses options and values.
    #    These are options in case you don't want to load the entire
    #    dataset, but only particular sets of projections.
    #    The possible arguments are:
    #         'sampling': type of sampling. default 'equidistant' Can be:
    #                'equidistant': equidistantly sample the entire set
    #                     of angles. 'num_angles' should be set for partial
    #                     loading of the data.
    #                'step': sample the entire set of projections every
    #                       'sampling_step' angles.
    #                'continous': Load the first 'num_angles' amount of
    #                           angles only.
    #
    #         'num_angles': Number of total angles to load. Default all of
    #                  them. Useful for 'equidistant' and 'continous' loading
    #
    #         'sampling_step': step to load when loading projections.
    #                 Default=1. Useful for 'step' loading.

    folder, geometry, angles = readXtekctGeometry(filepath)
    return loadNikonProjections(folder, geometry, angles, **kwargs)


def readXtekctGeometry(filepath):

    # Developed by A. Biguri and W. Sun
    # W. Sun edited on 06.10.2018 for 3D pro version 5.2.6809.15380 (2018)

    if filepath.endswith(".xtekct"):
        folder, ini = os.path.split(filepath)
    else:
        folder = filepath
        files = [file for file in os.listdir(folder) if file.endswith(".xtekct")]
        if not files:
            raise ValueError("No .xtekct file found in folder: " + folder)
        ini = files[0]

    cfg = ConfigParser()
    cfg.read(os.path.join(folder, ini))
    cfg = cfg["XTekCT"]

    geometry = Geometry()
    geometry.accuracy = 0.5

    ## Detector information
    # Number of pixel in the detector
    geometry.nDetector = numpy.array((int(cfg["DetectorPixelsX"]), int(cfg["DetectorPixelsY"])))
    # Size of pixels in the detector
    geometry.dDetector = numpy.array(
        (float(cfg["DetectorPixelSizeX"]), float(cfg["DetectorPixelSizeY"]))
    )
    # Total size of the detector
    geometry.sDetector = geometry.nDetector * geometry.dDetector

    ## Offset of the detector:
    geometry.offDetector = numpy.array(
        (float(cfg["DetectorOffsetX"]), float(cfg["DetectorOffsetY"]))
    )

    ## Image information
    # Number of voxels for the volume
    # the algos require these to be integers.
    geometry.nVoxel = numpy.array((int(cfg["VoxelsX"]), int(cfg["VoxelsY"]), int(cfg["VoxelsZ"])))
    # Size of each voxel
    geometry.dVoxel = numpy.array(
        (float(cfg["VoxelSizeX"]), float(cfg["VoxelSizeY"]), float(cfg["VoxelSizeZ"]))
    )
    # Size of the image in mm
    geometry.sVoxel = geometry.nVoxel * geometry.dVoxel
    geometry.offOrigin = numpy.array((0, 0, 0))

    #%% Global geometry
    geometry.DSO = float(cfg["SrcToObject"])
    geometry.DSD = float(cfg["SrcToDetector"])
    geometry.COR = -float(cfg["CentreOfRotationTop"])

    if geometry.COR == 0:
        print(
            "Centre of Rotation seems to be zero. Make sure that it is true and that the machine did not omit that information."
        )
    else:
        print(
            "TIGRE doesn't know if the sign of COR is the right one. Consider trying both and reporting to tigre.toolbox@gmail.com."
        )

    ## whitelevel
    geometry.whitelevel = float(cfg["WhiteLevel"])

    ## angles
    angles = []

    # It can be either an .ang  or .txt file
    # .ang
    files = [file for file in os.listdir(folder) if file.endswith(".ang")]
    if files:
        with open(os.path.join(folder, files[0]), "r") as file:
            # TODO: detect header lines automatically
            file.readline()
            for line in file:
                angles.append(math.radians(line.split("\t")[1]))
        return folder, geometry, numpy.array(angles)

    # .txt
    files = [file for file in os.listdir(folder) if file.endswith("_ctdata.txt")]
    if files:
        with open(os.path.join(folder, files[0]), "r") as file:
            # TODO: detect header lines automatically
            for i in range(3):
                file.readline()
            for line in file:
                angles.append(math.radians(float(line.split("\t")[1])))
        return folder, geometry, numpy.array(angles)

    print("File with definition of angles not found, estimating them from geometry info.")
    n_angles = int(cfg["Projections"])
    angle_step = float(cfg["AngularStep"])
    initial_angle = float(cfg["InitialAngle"])
    angles = numpy.arange(n_angles) * math.radians(angle_step) + math.radians(initial_angle)

    return folder, geometry, angles


def loadNikonProjections(folder, geometry, angles, **kwargs):
    # [projections, geometry, angles] = loadNikonProjections(filepath, geometry, angles, **kwargs)
    #    loads Nikon uCT machine projections
    #
    #    loadNikonData(filepath, geometry, angles) Loads a dataset given its FILEPATH,
    #       GEOMETRY and ANGLES (loaded from readXtekctGeometry())
    #       See NikonDataLoader() for additional options.
    #
    #% developed by A. Biguri and W. Sun 06.07.2020

    # parse inputs
    angles, indices = parse_inputs(geometry, angles, **kwargs)

    # load images
    files = sorted([file for file in os.listdir(folder) if file.lower().endswith(".tif")])

    image = Image.open(os.path.join(folder, files[indices[0]]))
    image = numpy.asarray(image).astype(numpy.float32)
    projections = numpy.zeros([len(indices), image.shape[0], image.shape[1]], dtype=numpy.single)
    projections[0, :, :] = -numpy.log(image / float(geometry.whitelevel))
    # Python uses zero-based indexing
    # if we start with index = 1 we get an error
    # use enumerate, it's cleaner
    print("Loading Nikon dataset: " + folder)
    for index, i in enumerate(tqdm(indices[1:]), 1):
        image = Image.open(os.path.join(folder, files[i]))
        image = numpy.asarray(image).astype(numpy.float32)
        projections[index, :, :] = -numpy.log(image / float(geometry.whitelevel))

    del geometry.whitelevel

    return numpy.asarray(projections), geometry, angles


def parse_inputs(geometry, angles, **kwargs):

    # TODO: warn user about invalid options or values
    sampling = kwargs["sampling"] if "sampling" in kwargs else "equidistant"
    nangles = int(kwargs["num_angles"]) if "num_angles" in kwargs else len(angles)
    step = int(kwargs["sampling_step"]) if "sampling_step" in kwargs else 1

    indices = numpy.arange(0, len(angles))

    if sampling == "equidistant":
        step = int(round(len(angles) / nangles))
        indices = indices[::step]
        angles = angles[::step]
    elif sampling == "continuous":
        indices = indices[:nangles]
        angles = angles[:nangles]
    elif sampling == "step":
        indices = indices[::step]
        angles = angles[::step]
    else:
        raise ValueError("Unknown sampling type: " + str(sampling))

    return angles, indices


if __name__ == "__main__":
    projections, geometry, angles = NikonDataLoader(
        "NikonDataLoaderTest/", sampling="continuous", num_angles=100
    )
    print("", projections.shape, "", geometry, "", angles, "", sep="\n")



"""
%--------------------------------------------------------------------------
% This file is part of the TIGRE Toolbox
%
% Copyright (c) 2015, University of Bath and
%                     CERN-European Organization for Nuclear Research
%                     All rights reserved.
%
% License:            Open Source under BSD.
%                     See the full license at
%                     https://github.com/CERN/TIGRE/license.txt
%
% Contact:            tigre.toolbox@gmail.com
% Codes:              https://github.com/CERN/TIGRE/
% Coded by:           MATLAB (original): Manasavee Lohvithee
                      Python: Reuben Lindroos
%--------------------------------------------------------------------------
"""
from __future__ import division

from functools import reduce

import numpy as np


def Measure_Quality(res_prev, res, QualMeasOpts):  # noqa: N803
    """
    :param res_prev: (np.ndarray)
        object being compared to
    :param res: (np.ndarray)
        true image
    :param QualMeasOpts: (str or listof(str))
        any of: 'RMSE', 'nRMSE' , 'CC', 'MSSIM', 'UQI', 'SSD'
    :return:
    """
    values = []
    # make it a list:
    for qual in QualMeasOpts:
        if "RMSE" == qual:
            N = np.prod(res_prev.shape)
            diff = res_prev - res
            values.append(np.sqrt(np.sum(diff ** 2) / N))
        if "nRMSE" == qual:
            N = reduce(lambda x, y: x * y, res_prev.shape)
            diff = res_prev - res
            values.append(
                (np.sqrt(np.sum(diff ** 2) / N) / ((0.00001 + np.sqrt(np.sum(res ** 2)) / N)))
            )
        if "CC" == qual:
            cc = np.corrcoef(res_prev.ravel(), res.ravel())
            values.append(cc[0, 1])

        if "MSSIM" == qual:
            N = np.prod(res_prev.shape)

            # Compute the mean pixel values of the two images

            mean_res_p = res_prev.mean()
            mean_res = res.mean()
            if mean_res == 0 and mean_res_p == 0:
                raise ValueError("Initialising with 0 matrix not valid")
            # Luminance Comparison

            K1 = 0.01  # K1 is a small constant <<1
            d = np.max(res_prev) - np.min(res_prev)  # dynamic range of the pixel values
            l = ((2 * mean_res * mean_res_p) + (K1 * d) ** 2) / (
                (mean_res_p ** 2) + (mean_res ** 2) + K1 * d ** 2
            )

            # Contrast comparison

            K2 = 0.02
            sres_p = res_prev.std()
            sres = res.std()

            c = ((2 * sres_p * sres) + (K2 * d) ** 2) / ((sres_p ** 2) + (sres ** 2) + K2 * d ** 2)

            # Structure comparison
            diffres_p = res_prev - mean_res_p
            diffres = res - mean_res
            delta = (1 / (N - 1)) * np.sum(diffres_p * diffres)
            s = (delta + (((K2 * d) ** 2)) / 2) / ((sres_p * sres) + ((K2 * d ** 2) / 2))

            values.append((1 / N) * l * c * s)
        if "UQI" == qual:
            N = np.prod(res_prev.shape)
            # Mean
            mean_res_p = np.mean(res_prev)
            mean_res = np.mean(res)

            # Variance
            varres_p = np.var(res_prev)
            varres = np.var(res)
            if mean_res == 0 and mean_res_p == 0:
                raise ValueError("Initialising with 0 matrix not valid")
            # Covariance
            cova = np.sum((res - mean_res) * (res_prev - mean_res_p)) / (N - 1)
            front = (2 * cova) / (varres + varres_p)
            back = (2 * mean_res * mean_res_p) / ((mean_res ** 2) + (mean_res_p ** 2))

            values.append(np.sum(front * back))
        if "SSD" == qual:
            values.append(np.sum((res_prev - res) ** 2))
    if len(values) == 1:
        return np.array(values[0])
    else:
        return np.array(values)
